################################################################################
#        Library for one-clock timed automata diagnozer simulation             #
#                                                                              #
#                       JAZIRI Samy - 01/04/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#                                                                              #
################################################################################

from TILib.timlib import *
from TILib.milib import IntervalFromSL, INFINITY
import TILib.ruilib as ruilib

from ATDLib.otaumata import OTAutomata

from random import randint
from itertools import product
from functools import reduce
#import sys

################################################################################
#### Class DiagOTA
################################################################################

class TripDOTA :
    '''
    '''
    # Class Attribute
    counter = 0
    
    # Constructor
    def __init__(self,states,guards,trans,init_state,e_trans, name = '') :
        ''' inter_pcal is a 3D matric indexed on states x states x guards 
        containing an array :
        inter_pcal[i] = the list of g' such that g' is a guard and  
        L_0^{guards[i]} \cap T^{guards[i],g'} \neq \emptyset

        parik_pcal is a 3D matric indexed on states x states x guards 
        containing an array of triple :
        parik_pcal[i] = the list of (g',G_{e,g},J_{guards[i],g',g,e})
        '''
        self.states = states; self.guards = guards; self.trans = trans;
        self.init_state = init_state
        self.silent = ''
                          
        if name :
            self.name = name 
        else :
            self.name = 'TDOTA #%i' % DiagOTA.counter
            TripDOTA.counter += 1
            
    # Methods
    def apply(self,label,tim) :
        ''' Apply the only transition t associated to label l on tim.
        the transition t is a array of a 2-element array of two arrays with 
        the form :
               q1         q2  ...       qn 
            [ 0: , id: ]           [ 0: , id: ]   
        q1   G11   G11'              .     .
        q2   G21   G21'              .     .
        .     .     .                .     .
        .     .     .                .     .
        .     .     .                .     .
        qn   Gn1   Gn1'              .     .

        with Gij the guard of the transition (qi,Gij,reset,qj) in the OTA
         and Gij' the guard of the transition (qi,Gij',id,qj) in the OTA
        Guards are equal to None if there are no transition between the states 
        The hole column j is None if no transition in the OTA leads to qj
        '''
        g = self.guards
        tim_tmp = [EMPTYFUTI(g) for x in range(0,len(self.states))]
        tim.apply_dec()
        
        for q,t in enumerate(self.trans[label]) :
            if t :
                for (l,guard) in enumerate(t[0]) :
                    if guard and tim[l] & guard : tim_tmp[q] = ZEROFUTI(g); break
                for (l,guard) in enumerate(t[1]) :
                    if guard : tim_tmp[q] | (tim[l] & guard)

        res = reduce(lambda x,y : bool(x) or bool(y), tim_tmp, False)
        if res :
            for q,futi in enumerate(tim_tmp) : tim[q] = futi
            self.e_closure(tim,0)
        return res

    ## Below we find the methods to comupte the epsilon closure
    def e_closure(self,tim,d) :
        states = self.states
        guards = self.guards
        
        def one_step_acc(old_tim) :
            new_tim = [EMPTYFUTI(g) for x in range(0,len(s))]
            new_tim = new_tim | old_tim
            for (q,lg,hg,op,l) in self.e_trans :
                for g in guards :
                    if g < hg and bool(old_tim[q][g]) :
                        int_hg = IntervalFromSL((milib.INFINITY,True),
                                                guards[hg])
                        if op :
                            new_tim[l][max(g,lg)] = (new_tim[l][max(g,lg)] |
                                                     (old_tim[q][g] & int_hg) )
                        else :
                            int_d = IntervalFromSL((-d,False),(0,True))
                            G = IntervalFromSL(guards[max(g,lg)],guards[hg])
                            new_tim[l][0] = (new_tim[l][0] |
                                             ( ((old_tim[q][g] & int_hg) - G) &
                                               int_d ) )
            return new_tim

        old_tim = tim
        new_tim = one_step_acc(tim)
        while new_tim != old_tim :
            old_tim = new_tim
            new_tim = one_step_acc(old_tim)

        tim = (new_tim + d).sem()
        

    def init_config(self) :
        init_futi = []
        for i in range(0,len(self.states)) :
            init_futi += ( [ZEROFUTI(self.guards)] if i == self.init_state
                           else [EMPTYFUTI(self.guards)] )
        return self.e_closure(TIMarking(self.states,self.guards, init_futi))
    
    def __call__(self,l,tim) :
        if type(l) == type('') :
            return self.apply(l,tim)
        else :
            self.e_closure(tim,d)
            
################################################################################
#### Function for TripDOTA construction from OTA
################################################################################

code = '''from Test_sample.{0} import ota
from TILib.milib import Interval, MultiInterval, INFINITY
from TILib.ruilib import RegUInterval
from ATDLib.trip_diagotaumata import TripDOTA

states = ota.states; guards = ota.guards; silent = ota.silent
init_state = ota.init_state; e_trans = ota.get_silent_trans()

trans = {1}

trip_dota = TripDOTA(states,guards,trans,init_state,e_trans,{2})
'''

def trip_dota_from_ota(ota,retres = False) :
    n = len(ota.states); m = len(ota.guards)
    
    trans = { l : [ [ [None]*n, [None]*n ] for q in ota.states ]
              for l in ota.trans }

    for l in ota.trans :
        for t in ota.trans[l] :
            trans[l][t[4]][int(t[3])][t[0]] = (t[1],t[2])
            
    if retres :
        return TripDOTA(ota.states,ota.guards,trans,ota.init_states,
                        ota.get_silent_trans(),
                        'trip_diag_' + ota.name)
        
    wota = open('Test_sample/trip_dota_' + ota.name + '.py','w')                       
    wota.write(code.format(ota.name,
                           trans,
                           repr('diag_' + ota.name)) )
