################################################################################
#              Library for regular union of intervals calculus                 #
#                                                                              #
#                       JAZIRI Samy - 30/03/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#  - We represent a regular union of intervals as a 5-tuple : (I,J,M,p,dec)    #
#    where I is a multi-interval                                               #
#          J is a multi-interval such that J is a a subset of ]-p,0]           #
#          p is a positive integer                                             #
#          M is a positive integer such that I is a subset of ]-M,inf[         #
#          dec is a positive float                                             #
#    This tuple represent the set                                              #
#          ( I \cup \bigcup_{k = 0}^{inf} (J-M)-kp ) + d                       #
#                                                                              #
################################################################################

import TILib.milib as milib
from TILib.milib import bigor, Interval, IntervalFromSL

from decimal import Decimal

INFINITY = Decimal('inf')

################################################################################
#### Class RegUInterval
################################################################################

class RegUInterval :
    ''' 
    Regular Union of Intervals
    Mutable
    
    Note : 
      We implement a regular union of intervals as a 5-tuple : (I,J,p,M,d) 
          where I is a multi-interval                                         
          J is a multi-interval such that J is a a subset of ]-p,0]           
          M is a positive integer such that I is a subset of ]-M,inf[         
          p is a positive integer                                             
          d is a positive float                                             
       
    Instantiate the emptyset by default.

    Args :
       I,J,p,M,d as described above

    Attributes :
       Ibase,Irep,period,Irep_dec,dec : as described above

    %%TODO : Implement union

    '''

    #Constructor
    def __init__(self,Ibase = milib.EMPTYSET, Irep = milib.EMPTYSET,
                 period = 0, Irep_dec = 0, dec = 0, precision = '.2') :
        if not Irep or not period : 
            self.Ibase = Ibase | (Irep-Irep_dec)
            self.Irep = milib.EMPTYSET
            self.period = 1
            self.Irep_dec = 0
        else :
            self.Ibase = Ibase
            self.Irep = Irep
            self.period = period
            self.Irep_dec = Irep_dec
            
        self.dec = dec
        self.pf = '%' + precision + 'f'
        self.adapt_Irep2period()

    # Representation
    def __str__(self) :
        ''' Return a string with the representation of self '''
        if not self : return 'EMPTYSET'
        res = str(self.Ibase)
        if self.Irep :
            sJ = str(self.Irep);
            if self.Irep_dec : sJ = sJ + ' - ' + self.pf % self.Irep_dec
            res = ('( ' + res + ' )' +' u U_{k in N} ( ' + sJ +
                   ' -k' + self.pf % self.period + ' ) ' )
        if self.dec : res = '[ ' + res + ' ] + ' + self.pf % self.dec
        return res

    def __repr__(self) :
        ''' Return a string with the code representing self '''
        return ','.join(['RegUInterval(' + repr(self.Ibase),
                        repr(self.Irep), repr(self.period),
                        repr(self.Irep_dec),repr(self.dec),
                         repr(self.pf)+')'])
                        
    
    # Methods
    ## Informations
    def vol(self) :
        ''' Return the volume of Ibase and of Irep '''
        return (self.Ibase.vol(),self.Irep.vol())

    def sup(self) :
        ''' Return the sup of self '''
        return (self.Ibase.sup() + self.dec if self.Ibase
                else self.Irep.sup()-self.Irep_dec+self.dec )

    def sup_is_max(self) :
        ''' Return boolean saying if the sup of self is a max : 
        True if its a max, False otherwise '''
        boo = self.Ibase.sup_is_max() if self.Ibase else self.Irep.sup_is_max() 
        return boo

    def inf(self) :
        ''' Return the inf of (Irep - Irep_dec) '''
        return self.Irep.inf() -self.Irep_dec        

    ## Modifications
    def apply_dec(self,addpad = 0) :
        ''' Effectively apply the paddings (+ an additional padding addpad). 
        Irep = Irep - Irep_dec + dec
        Ibase = Ibase + dec '''
        self.dec += addpad
        self.Ibase = self.Ibase + self.dec
        self.Irep_dec -= self.dec
        self.dec = 0

    def new_rdec(self,new_rdec) :
        ''' Change the Irep_dec and adapt the Ibase and Irep to verify the 
        constraints. Return a new equivalent RUI. '''
        if not self.Irep : return RegUInterval(self.Ibase,dec = self.dec)
        diff_dec = new_rdec - self.Irep_dec; assert diff_dec >= 0
        quot = int(diff_dec // self.period); rem = diff_dec - self.period*quot
        int_ptorem = Interval(']-'+str(self.period)+',-'+ str(rem)+']')
        int_remto0 = Interval(']-'+str(rem)+',0]') if rem else milib.EMPTYSET
        int_remtonewd = (Interval(']-'+str(new_rdec)+',-'+ str(new_rdec-rem)+']')
                         if rem else milib.EMPTYSET)
        args = { 'period' : self.period, 'Irep_dec' : new_rdec, 'dec' : self.dec,
                 'Ibase' : bigor( [self.Ibase] +
                                  [(self.Irep - self.Irep_dec - k*self.period)
                                   for k in range(0,quot) ] +
                                  [(self.Irep & int_remto0) - new_rdec + rem ] ),
                 'Irep' : ( ( (self.Irep & int_ptorem) + rem ) |
                            ( (self.Irep & int_remto0) - self.period + rem ) ) }
        return RegUInterval(**args)

    def new_mulperiod(self,n) :
        ''' Enlarge the period from p to n*p and adapt the Irep to verify the 
        constraints. Return a new equivalent RUI. '''
        assert n > 0
        if not self.Irep : return RegUInterval(self.Ibase,dec = self.dec)
        args = { 'Ibase' : self.Ibase, 'Irep_dec' : self.Irep_dec,
                 'dec' : self.dec, 'period' : n*self.period,
                 'Irep' : bigor([self.Irep - k*self.period
                                 for k in range(0,n)]) }
        return RegUInterval(**args)

    def adapt_Irep2period(self) :
        if not self.Irep : return
        self.apply_dec()
        
        if 0 < self.Irep.sup() :
            self.Irep_dec += self.Irep.sup()
            self.Irep = self.Irep - self.Irep.sup()
            self.adapt_Irep2period()
        elif (self.Irep.inf(),self.Irep.inf_is_min()) < (-self.period,True) :
            n = int( - self.Irep.inf() // self.period ) + 1
            pto0 = IntervalFromSL((-self.period,True),(0,True))
            npto0 = IntervalFromSL((-(n-1)*self.period,True),(0,True))
            self.Ibase = bigor( [self.Ibase] +
                                [( (self.Irep - k*self.period) & npto0 )
                                 - self.Irep_dec for k in range(0,n) ] )
            self.Irep_dec += (n-1)*self.period
            self.Irep = bigor( [ (self.Irep + i*self.period) & pto0
                                 for i in range(0,n) ] )

            if self.Irep == pto0 :
                rdec2inf = IntervalFromSL((-INFINITY,True),(-self.Irep_dec,True))
                self.Ibase = rdec2inf | self.Ibase 
                self.Irep = milib.EMPTYSET
                self.period = 0
                self.Irep_dec = 0
        
    # Operator Overloading
    #### Unary  Operators
    def __bool__(self) :
        ''' Is empty self '''
        return bool(self.Ibase) or bool(self.Irep)
    
    #### Set Operators
    def __and__(self,mi) :
        ''' Return the intersection of self and mi ( a multi-interval ).
        Complexity if mi is an interval : O( log( |I| + n|J| )
        with n = (-( inf(mi) + Irep_dec) / period ) + 1 
        
        If mi is a multiinterval, multiply by the size of mi '''
        if not self or not mi : return EMPTYSET #RegUInterval.EMPTYSET
                
        mi = mi - self.dec # We take advantage of I+d \cap J = (I \cap J-d) + d

        if not self.Irep : return RegUInterval(self.Ibase & mi, dec = self.dec)

        args = { 'dec' : self.dec }
        
        if mi.inf() == -INFINITY : #In this case only we'll return a RUI
            lim = (mi[0].sup(),True)
            args['Irep'] = self.Irep; args['period'] = self.period;
        else : lim = (mi.inf(),False)

        if lim >= (-self.Irep_dec,True) : args['Ibase'] = self.Ibase & mi
        else :
            n = int( -(lim[0] + self.Irep_dec) // self.period ) + 1
            args['Ibase'] = bigor( [self.Ibase] +
                                   [(self.Irep - self.Irep_dec - k*self.period)
                                    for k in range(0,n) ] ) & mi

            # Used only if mi.inf() == -INFINITY
            args['Irep_dec'] = self.Irep_dec + n*self.period 

        return RegUInterval(**args)

    #### Arithmetic
    def __add__(self,I) :
        ''' Add a float d to self. In place change and return self.
        Do nothing on EMPTYSET, EMPTYSET have no possible in place changes.
        Complexity : O(1) '''
        
        if self :
            if (type(I) == type(0) or
                type(I) == type(0.0) or
                type(I) == type(Decimal(0)) ) :
                self.dec += I
            else :
                self.apply_dec()

                if I.inf() == -INFINITY :
                    return RegUInterval(I+self.Ibase)

                args = { 'Ibase' : I + self.Ibase,
                         'Irep' : I + self.Irep,
                         'Irep_dec' : self.Irep_dec,
                         'period' : self.period }
                return RegUInterval(**args)
            
        return self

    def __sub__(self,I) :
        return self.Ibase - I

    def __eq__(self,other) :
        self.apply_dec(); other.apply_dec()
        return (self.Ibase == other.Ibase)
    ''' and
                self.Irep == other.Irep and
                self.period == other.period and
                self.Irep == other.Irep and
                self.dec == other.dec )'''

    def __neq__(self,other) :
        return not (self == other)
    
    def __or__(self,other) :
        ''' Temporary or implemented, works on multi-intervals only '''

        if not self.Irep :
            if not other.Irep : return RegUInterval(self.Ibase | other.Ibase)
            else :
                other.apply_dec()
                if self.Ibase.inf() == -INFINITY :
                    new_d = -self.Ibase.sl[1][0] 
                    if new_d > other.Irep_dec :
                        return RegUInterval(self.Ibase | 
                                            other.new_rdec(new_d+1))
                    else :
                        return RegUInterval(self.Ibase | other.Ibase)
                
                if -self.Ibase.inf() > other.Irep_dec :
                    rui2 = other.new_rdec(-self.Ibase.inf()+1)
                else : rui2 = other
                args = { 'Ibase' : self.Ibase | rui2.Ibase,
                         'Irep' : rui2.Irep,
                         'Irep_dec' : rui2.Irep_dec,
                         'period' : rui2.period }
        elif not other.Irep :
            self.apply_dec()
            if other.Ibase.inf() == -INFINITY :
                    new_d = -other.Ibase.sl[1][0] -1
                    if new_d > self.Irep_dec :
                        return RegUInterval(other.Ibase | 
                                            self.new_rdec(new_d+1))
                    else :
                        return RegUInterval(self.Ibase | other.Ibase)

            if -other.Ibase.inf() > self.Irep_dec :
                rui2 = self.new_rdec(-other.Ibase.inf()+1)
            else : rui2 = self
            args = { 'Ibase' : other.Ibase | rui2.Ibase,
                     'Irep' : rui2.Irep,
                     'Irep_dec' : rui2.Irep_dec,
                     'period' : rui2.period }
        else :
            self.apply_dec(); other.apply_dec()
            if self.Irep_dec < other.Irep_dec : rui1,rui2 = other,self
            else : rui1,rui2 = self,other
            
            rui2 = rui2.new_rdec(rui1.Irep_dec)
            lcmp1p2 = lcm(rui1.period,rui2.period)
            n1 = int(lcmp1p2 // rui1.period) ; n2 = int(lcmp1p2 // rui2.period)
            rui1 = rui1.new_mulperiod(n1); rui2 = rui2.new_mulperiod(n2) 
            
            args = { 'Ibase' : rui1.Ibase | rui2.Ibase,
                     'Irep' : rui1.Irep | rui2.Irep,
                     'Irep_dec' : rui1.Irep_dec,
                     'period' : lcmp1p2 }

        return RegUInterval(**args)

################################################################################
#### GCD and LCM Function
################################################################################
    
def gcd(a,b):
    '''Compute the greatest common divisor of a and b'''
    while b > 0:
        a, b = b, a % b
    return a
    
def lcm(a, b):
    '''Compute the lowest common multiple of a and b'''
    return int(a * b // gcd(a, b))
    
################################################################################
#### Running Tests
################################################################################

EMPTYSET = RegUInterval()
def ZERO() : return  RegUInterval(milib.ZERO)

if __name__ == '__main__' :
    from random import sample, randint, uniform
    import numpy
    numpy.set_printoptions(precision=2)

    NUMBER = 6
    SIZE = 5
    VERIFNUMBER = 10000
    population = ([(x / 100,True) for x in range(SIZE*1000)] +
                  [(x / 100,False) for x in range(SIZE*1000)] ) 
    [U,V,W,
     U1,V1,W1] = [ milib.MultiInterval(sorted(sample(population,randint(0,SIZE)*2)))
               for x in range(NUMBER) ]
    V = milib.Interval('[0,0]')-V; V1 = milib.Interval('[0,0]')-V1
    U = U - 20; U1 = U1 - 20
    if not V : p1 = 0
    else : p1 = -int(V.inf())+2
    if not U : M1 = 0
    else : M1 = -min(U.inf()-1,0)
    W = milib.Interval('[-30,-30]')-W
    W2 = milib.Interval(']-inf,-30[')
    R = RegUInterval(U,V,p1,M1,0)
    if not V1 : p2 = 0
    else : p2 = -int(V1.inf())+2
    if not U1 : M2 = 0
    else : M2 = -min(U1.inf()-1,0)
    R1 = RegUInterval(U1,V1,p2,M2,0)
    
    print(R,end='\n\n\n')
    #print(R.new_rdec(R.Irep_dec+R.period+(R.period/3)),end='\n\n')
    print(R1,end='\n\n\n')
    #print(R1.new_rdec(R1.Irep_dec+R1.period+(R1.period/3)),end= '\n\n')
    RU = R | R1
    #print(RU,end='\n\n\n')
    RA = R + W
    print(W,end='\n\n\n')
    print(RA,end='\n\n\n')
    
    def bel(x,U) :
        ''' Return True if x belongs (set membership) to the set U. 
        This is done by testing U \cap [x,x] == EMPTYSET '''
        return bool(U & Interval('[{0},{0}]'.format(x)))
'''
    count1 = 0
    valuex = 0
    for j in range(0,VERIFNUMBER) :
        x = uniform(0,SIZE*10)
        if randint(0,1) : x = -x
        valuex += x
        count1 += 1 if bel(x,RU) else 0
        if (bel(x,R) or bel(x,R1)) != bel(x,RU) :
            print( (
                    'X == Error UnM on RUIs {0} and {1} ' +
                    '\twith element {2}').format(R,R1,x) )
            break;
    else :
        print('] %i tests succeeded' % VERIFNUMBER)
        print('O == Union correction test succeeded\n')
        print('Average x in U \cap V is %i/%i \n Average value of x is %i' %
              (count1, VERIFNUMBER, valuex // VERIFNUMBER) )


'''    
#    print(W)
#   print(R & W)
#   print(W2)
#   print(R)
#   R2 = (R & W2)
#   R2.apply_dec()
#   print(R2)
    
