################################################################################
#              Library for timed interval markings calculus                    #
#                                                                              #
#                       JAZIRI Samy - 30/03/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#  - We assume that when loading the automatom we get :                        #
#      - a list guards of the semi-lines associated to each bounds             #
#        (low or high) of guards in the automaton. This list should be sorted  #
#      - a list states of states of the automaton.                             #
#  - We represent a finite union of timed interval by a list of regular        #
#    unions of interval. At index i we represent the r.u.i. associated to      #
#    semi-line guards[i].                                                      #
#    Such a list L represent the timed interval :                              #
#      I = \bigsqcup_{ i = 0 }^{ len(lbguards) } L[i] \sqcap lbguards[i]       #
#  - A timed interval marking is again a list of f.u.t.i. At index i we        #
#    represent the f.u.t.i. associated to state state[i]                       #
#                                                                              #
################################################################################

from TILib.ruilib import *
from TILib.milib import Interval, IntervalFromSL

from functools import reduce

################################################################################
#### Class FUTInterval
################################################################################

class FUTInterval :
    ''' 
    Finite Union of Timed Interval
    Mutable Sequence

    The object is meant to undergo in place changes.
    
    Note : 
      A finite union of timed interval is represented by a list of regular 
      union of interval, one for each lower bound of guards appearing in the 
      automaton we want to diagnoze. Hence we keep in reality two lists :
        - guards : an immutable sorted list with for each index the semi-line
                     associated ( a lower bound of a guard ). 
                     First element is always (0,False) and the last element 
                     (INFINITY,False)
        - ruis     : a list of RegUIntervals of the size of guards - 1. 
                     The RegUInterval at index i is associated to semi-line 
                     at i in guards.
       
    Instantiate the emptyset by default.

    Args :
       guards, ruis as described above
       dec : a general padding on all r.u.i

    Attributes :
       guards, ruis as described above
       dec : a general padding on all r.u.i

    '''

    #Constructor
    def __init__(self,guards,ruis = None, dec = 0) :
        assert ( len(guards) > 1 and guards[0] == (0,False) and
                guards[-1] == (INFINITY,False) ) 
        self.ruis = ruis if ruis else [EMPTYSET for x in range(1,len(guards))]
        self.guards = guards
        self.dec = dec

    # Representation
    def __repr__(self) :
        ''' Return a string with the representation of self '''
        res = '\n |_| '.join([ '(' + str(I) + ') n ' + str(IntervalFromSL(g))
                               for I,g in zip(self.ruis,self.guards) if I ])
        if self.dec : res = '[ ' + res + '] + ' + str(self.dec)
        return res
                

    # Method
    def interGuard(self,i1,i2) :
        ''' If i1 and i2 are two index of guards, return the intersection
        between self.sem() and guards[i1] \cap \overline{ guards[i2] } '''

        def IFromIndex(j1,j2) :
            return IntervalFromSL(self.guards[j1], self.guards[j2]) - self.dec
        
        guard = IFromIndex( i1, i2 )
        
        res = bigor([self.ruis[i] & guard for i in range(0,i1+1)] +
                    [self.ruis[i] & IFromIndex(i,i2) for i in range(i1+1,i2)])
        return res + self.dec
        
    def sem(self) :
        ''' Return the semantic of the futi, i.e. 
        \bigcup_{i = 0}^{ len(guards) } ruis[i] \cap guards[i] 
        We use method interGuard with param 0, len(guards) to get it'''
        
        return self.interGuard(0,len(self.guards)-1) + self.dec

    def propagate_dec(self) :
        ''' Convert the general padding stored in the FUTInterval into a padding
        for each RegUInterval of self.ruis '''
        for x in map(lambda x : x+self.dec,self.ruis) : pass
        self.dec = 0
    
    def apply_dec(self,addpad = 0) :
        ''' Effectively apply the paddings on all r.u.i (c.f. apply_dec in 
        module ruilib) + an additional padding addpad '''
        self.dec += addpad
        for x in map(lambda x : x.apply_dec(self.dec), self.ruis) : pass
        self.dec = 0

    def comp_time(self) :
        ''' Compute the remaining time before self.sem() becomes not empty '''
        return reduce( lambda x,y : min(x, y),
                       [ (g[0] - (I.sup() + self.dec), not g[1] & I.sup_is_max()) 
                         for I,g in filter( lambda x : bool(x[0]),
                                            zip(self.ruis,self.guards) ) ],
                       (INFINITY,False))
            
        
    #Operator Overloading
    def __bool__(self) :
        ''' Return True if and only if not emptyset '''
        return reduce(lambda x,y : bool(x) or bool(y), self.ruis, False)

    def __getitem__(self,i) :
        ''' Return the RUI at i '''
        return self.ruis[i]

    def __setitem__(self,i,val) :
        ''' Change the RUI at i with val '''
        self.ruis[i] = val
    
    def __and__(self,guard) :
        ''' We just consider that guard is a couple of indexes and call 
        interGuard. '''
        return self.interGuard(guard[0],guard[1])

    def __add__(self,d) :
        ''' We implement for now just addition by a float : self + d.
        Make in place changes and return self.'''
        self.dec += d
        return self

    def __sub__(self,d) :
        ''' We implement for now just subtraction by a float : self - d.
        Make in place changes and return self.'''
        self.dec -= d
        return self

    def __or__(self,other) :
        ''' We implement for now just union with an interval 
        Make in place changes and return self '''
        if (isinstance(other,RegUInterval) or
            isinstance(other,milib.MultiInterval) ) :
            self.ruis[0] = self.ruis[0] | other
        else :
            for i in range(0,len(self.guards)-1) : 
                self.ruis[i] = self.ruis[i] | other.ruis[i]


    def __eq__(self,other) :
        for r1,r2 in zip(self,other) :
            if r1 != r2 : return False
        return True

    def __neq__(self,other) :
        return not (self == other)
    
def EMPTYFUTI(guards) : return FUTInterval(guards)
def ZEROFUTI(guards) :
    return FUTInterval(guards,[ZERO()] +
                       [EMPTYSET for x in range(1,len(guards))])

################################################################################
#### Class TIMarking
################################################################################

class TIMarking :
    ''' 
    Timed Interval Marking
    Mutable Sequence
    
    Note : 
      A timed interval marking is represented by a FUTInterval for each state
      of the automaton we want to diagnoze. Hence in reality we keep two lists
        - state : an immutable sorted list with for each index the state
                  associated. 
        - futis : a list of FUTIntervals of the same size of state. 
                  The FUTInterval at index i is associated to state 
                  at i in state.
       
    Instantiate the empty marking by default.

    Args :
       state, futis as described above
       dec : a general padding for all futis.
    
    Attributes :
       state, futis as described above
       dec : a general padding for all futis.

    '''

    #Constructor
    def __init__(self,state,guards,futis = None,dec = 0) :
        assert ( len(guards) > 1 and guards[0] == (0,False) and
                guards[-1] == (INFINITY,False) ) 
        self.futis = futis if futis else [EMPTYFUTI(guards)
                                          for x in range(1,len(state))]
        self.state = state
        self.guards = guards
        self.dec = dec
        
    # Representation
    def __repr__(self) :
        ''' Return a string with the representation of self '''
        
        return (reduce( lambda x,y : (x + '\nState : ' + self.state[y[0]] + ':\n'
                                     + '-' * 15 + '\n' + str(y[1]) + '\n') ,
                       enumerate(self.futis),'' ) +
                '-'*15 + '\nGeneral Padding on all states : ' + str(self.dec))

    # Methods
    def propagate_dec(self) :
        ''' Convert the general padding stored in the TIMarking into a padding
        for each RegUInterval of self.ruis '''
        for x in map(lambda x : x+self.dec,self.futis) : pass
        self.dec = 0
    
    def apply_dec(self,addpad = 0) :
        ''' Effectively apply the paddings on all r.u.i (c.f. apply_dec in 
        module ruilib) + an additional padding addpad '''
        self.dec += addpad
        for x in map(lambda x : x.apply_dec(self.dec), self.futis) : pass
        self.dec = 0

    def sem(self) :
        ''' Return the semantic of TIMarking with for each state the sem() of
        the FUTInterval associated '''
        self.apply_dec()
        return TIMarking(self.state,self.guards,
                         [FUTInterval(self.guards,
                                      ([F.sem()] +
                                       [EMPTYSET for x in range(1,
                                                                len(self.guards))]))
                          for F in self.futis])

    def affect_sem(self) :
        ''' Change a TIMarking for its semantic '''
        self.apply_dec()
        self.futis = [FUTInterval(self.guards,
                                  ([F.sem()] +
                                   [EMPTYSET for x in range(1,
                                                            len(self.guards))]))
                      for F in self.futis]
    
    def reachable_states(self,comp_time = True) :
        ''' Return a couple, with as left member the set of reached states
        and as right member the set reachable state if comp_time = False or 
        the dictionnary with reachable states as keys and the time before we
        reach them as values '''

        reached = {}
        reachable = {}

        for i in range(0,len(self.state)) :
            self.futis[i] + self.dec
            if self.futis[i].sem() : reached[self.state[i]] = 0
            elif self.futis[i] :
                reachable[self.state[i]] = (self.futis[i].comp_time()
                                            if comp_time else 'Unknown')
        self.dec = 0
        return (reached,reachable)
    
    # Overloading Operators
    def __getitem__(self,i) :
        self.apply_dec()
        return self.futis[i]

    def __setitem__(self,i,val) :
        self.futis[i] = val + (-self.dec)
    
    def __add__(self,d) :
        ''' We implement for now just addition by a float : self + other 
        return self. '''
        self.dec += d
        return self

    def __eq__(self,other) :
        for f1,f2 in zip(self,other) :
            if f1 != f2 : return False
        return True

    def __neq__(self,other) :
        return not (self == other)

    def __or__(self,other) :
        for f1,f2 in zip(self,other) :
            f1 | f2
    
################################################################################
#### Running Tests
################################################################################
                                    
if __name__ == '__main__' :
    from random import sample, randint
    
    NUMBER = 8
    SIZE = 5
    VERIFNUMBER = 1000
    population = ([(x / 100,True) for x in range(SIZE*1000)] +
                  [(x / 100,False) for x in range(SIZE*1000)] ) 
    LI = [ milib.MultiInterval(sorted(sample(population,randint(0,SIZE)*2)))
           for x in range(NUMBER) ]
    LI2 = [ milib.MultiInterval(sorted(sample(population,randint(0,SIZE)*2)))
           for x in range(NUMBER) ]

    guards = [(0,False),(110,True),(120,False),(130,False),(140,True),(INFINITY,False)]
    print(ZEROFUTI(guards))

    def def_args(U,V) :
        V = milib.Interval('[0,0]')-V
        args = { 'Ibase' : U, 'Irep' : V }
        if not V : args['period'] = 0
        else : args['period'] = -int(V.inf())+2
        if not U : args['Irep_dec'] = 0
        else : args['Irep_dec'] = -min(U.inf()-1,0)
        return args
    
#    RUI1 = [RegUInterval(**def_args(U,V)) for U,V in zip(LI[::2],LI[1::2])]
#    RUI2 = [RegUInterval(**def_args(U,V)) for U,V in zip(LI2[::2],LI2[1::2])]

    RUI1 = [EMPTYSET] + [RegUInterval(U) for U in LI[::2]]
    RUI2 = [EMPTYSET] + [RegUInterval(U) for U in LI2[::2]]

    T1 = FUTInterval(guards,RUI1)
    T2 = FUTInterval(guards,RUI2)
    print(T1,end='\n' + '-'*30+'\n')
    print(T2,end='\n' + '-'*30+'\n')
    #print(T1)
    #print(T.sem())
    #print(T.interGuard(1,2), T.interGuard(1,3))  
    T1+5
    T1.apply_dec()
    print(T1,end='\n' + '-'*30+'\n')
    #print(EMPTYFUTI(guards))

    state = ['q0','q1']
    M = TIMarking(state,guards,[T1,T2])
    M + 10
    print(M)
    M.apply_dec()
    print(M)
    print(M.sem())
    rs = M.reachable_states()
    print(rs)
    print(M + rs[1]['q0'][0] + 1)
    print(M.reachable_states())
    print(M)
