################################################################################
#        Library for one-clock timed automata diagnozer simulation             #
#                                                                              #
#                       JAZIRI Samy - 01/04/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#                                                                              #
################################################################################

from TILib.timlib import *
from TILib.milib import IntervalFromSL, INFINITY
import TILib.ruilib as ruilib

from ATDLib.otaumata import OTAutomata

from random import randint
from itertools import product
from functools import reduce
#import sys

################################################################################
#### Class DiagOTA
################################################################################

class DiagOTA :
    '''
    '''
    # Class Attribute
    counter = 0
    
    # Constructor
    def __init__(self,states,guards,trans,init_state,
                 inter_pcal = [], parik_pcal = [], name = '') :
        ''' inter_pcal is a 3D matric indexed on states x states x guards 
        containing an array :
        inter_pcal[i] = the list of g' such that g' is a guard and  
        L_0^{guards[i]} \cap T^{guards[i],g'} \neq \emptyset

        parik_pcal is a 3D matric indexed on states x states x guards 
        containing an array of triple :
        parik_pcal[i] = the list of (g',G_{e,g},J_{guards[i],g',g,e})
        '''
        self.states = states; self.guards = guards; self.trans = trans;
        self.init_state = init_state
        self.silent = ''
        if inter_pcal : self.inter_pcal = inter_pcal
        else :
            self.inter_pcal = [ [] for x in range(0, len(states)) ]
            for l in range(0,len(states)) :
                for q in range(0,len(states)) :
                    self.inter_pcal[l] += [ [ [milib.REALS] if q == l else [] ] +
                                            [ [] ]*(len(guards)-2) ]
                           
        self.parik_pcal = (parik_pcal if parik_pcal else
                           [ [ [ [] ] * len(guards) ] * len(states)
                           ] * len(states) )
                          
        if name :
            self.name = name 
        else :
            self.name = 'DOTA #%i' % DiagOTA.counter
            DiagOTA.counter += 1
            
    # Methods
    def apply(self,label,tim) :
        ''' Apply the only transition t associated to label l on tim.
        the transition t is a array of a 2-element array of two arrays with 
        the form :
               q1         q2  ...       qn 
            [ 0: , id: ]           [ 0: , id: ]   
        q1   G11   G11'              .     .
        q2   G21   G21'              .     .
        .     .     .                .     .
        .     .     .                .     .
        .     .     .                .     .
        qn   Gn1   Gn1'              .     .

        with Gij the guard of the transition (qi,Gij,reset,qj) in the OTA
         and Gij' the guard of the transition (qi,Gij',id,qj) in the OTA
        Guards are equal to None if there are no transition between the states 
        The hole column j is None if no transition in the OTA leads to qj
        '''
        g = self.guards
        tim_tmp = [EMPTYFUTI(g) for x in range(0,len(self.states))]
        tim.apply_dec()
        
        for q,t in enumerate(self.trans[label]) :
            if t :
                for (l,guard) in enumerate(t[0]) :
                    if guard and tim[l] & guard : tim_tmp[q] = ZEROFUTI(g); break
                for (l,guard) in enumerate(t[1]) :
                    if guard : tim_tmp[q] | (tim[l] & guard)

        res = reduce(lambda x,y : bool(x) or bool(y), tim_tmp, False)
        if res :
            for q,futi in enumerate(tim_tmp) : tim[q] = futi
            self.e_closure(tim)
        return res

    ## Below we find the methods to comupte the epsilon closure
    def eta(self,I,q,l,nu) :
        ''' Compute \eta(I,0,[ L ]_{0}^{nu}) '''
        G = self.guards
        n = len(G); m = len(self.trans)
        inter = bigor([ I & g for g in self.inter_pcal[q][l][nu] ])
        parikU = bigor([ T[2] + ( ( (I & T[0]) - T[1] ) & milib.NREALS )
                         for T in self.parik_pcal[q][l][nu] ])
        return inter | parikU

    def e_closure(self,tim) :
        s = self.states
        g = self.guards
        tim_tmp = [EMPTYFUTI(g) for x in range(0,len(s))]
        for l in range(0,len(s)) :
            for q in range(0,len(s)) :
                ruis_tmp = []
                for nu in range(0,len(g)-1) :
                    ruis_tmp.append(self.eta(tim[q][0],q,l,nu))
                tim_tmp[l] | FUTInterval(g,ruis_tmp)
        for q,futi in enumerate(tim_tmp) : tim[q] = futi
        return tim

    def init_config(self) :
        init_futi = []
        for i in range(0,len(self.states)) :
            init_futi += ( [ZEROFUTI(self.guards)] if i == self.init_state
                           else [EMPTYFUTI(self.guards)] )
        return self.e_closure(TIMarking(self.states,self.guards, init_futi))
    
    def __call__(self,l,tim) :
        if type(l) == type('') :
            return self.apply(l,tim)
        else :
            tim + d
            
################################################################################
#### Function for DiagOTA construction from OTA
################################################################################

def inter_pcal_from_ota(ota,silent = '') :
    ''' Take an one-clock timed automata, ota, and consider silent as a 
    silent transition (if none is specified, we'll take the default silent 
    transition of ota), then compute the inter_pcal to be passed as argument to
    the diagnozer of ota '''
    
    states = ota.states; guards = ota.guards; n = len(states); m = len(guards)
    indexes,access_g = ota.e_access_graph(silent)
    
    ''' For all q,l,nu find for all g if there is a path in access_g between 
    q,0,inf and l,nu,g with only identity transition, if so add g at position 
    q,l,nu in inter_pcal '''

    for i,row in enumerate(access_g) :
        for j,t_list in enumerate(row) :
            access_g[i][j] = list(filter(lambda t: t[2][3],t_list))
            
    inter_pcal = [ [ [ [] for nu in guards ] for l in states ] for q in states ]
    
    for q,l,nu in product(range(n),range(n),range(m-1)) :
        index_init = indexes.index((q,0,m-1)) 
        lifo = [index_init]
        marqued_indexes = [i == index_init for i in range(0,len(indexes))]

        while lifo :
            node = lifo.pop()
            
            if indexes[node][:2] == (l,nu) :
                H = IntervalFromSL((-milib.INFINITY,True),
                                   ota.guards[indexes[node][2]])
                inter_pcal[q][l][nu].append(H)

            is_unmarq_neb = (lambda nei : bool(access_g[node][nei]) and
                             not marqued_indexes[nei])
                                      
            for neighboor in filter(is_unmarq_neb, range(len(indexes))) :
                lifo.append(neighboor)
                marqued_indexes[neighboor] = True

    return inter_pcal

def parikh_image(ota,silent='') :
    states = ota.states; guards = ota.guards; n = len(states); m = len(guards)
    e_trans = ota.get_silent_trans(silent)
    indexes,access_g = ota.e_access_graph(silent)

    ''' Suppression of the information of higher bounds guards '''
    
    new_indexes = [ (q,low_g)
                    for q in range(n)
                    for low_g in range(m-1) ] 

    def tlist_from_indexes(i,j,hg,hg1) :
        old_i = indexes.index((i[0],i[1],hg))
        old_j = indexes.index((j[0],j[1],hg1))
        return access_g[old_i][old_j]
                                                      
    new_access_g = [ [ list({ (t[0],t[2],t[3])
                         for hg,hg1 in product(range(1,m),repeat = 2)
                         for t in tlist_from_indexes(s_i,s_o,hg,hg1) })
                       for s_o in new_indexes ]
                     for s_i in new_indexes ]

    indexes,access_g = new_indexes,new_access_g

    ''' Implementation de l'algorithme de McNaughton & Yamada adapté pour 
    calculer l'image de Parikh de ota pour chaque paire d'état q,0 et l,nu '''
    
    trans_list = [ (lg,t,lg1)
                   for (lg,lg1) in product(range(m-1),repeat = 2)
                   for t in e_trans ]

    index_init = lambda q : indexes.index((q,0))
    
    def new_tuple_trans(trans) :
        return tuple([ int( i == trans_list.index(trans) )
                       for i in range(len(trans_list)) ] )
    
    vec_zero = (0,)*len(trans_list)

    def check_tuple_is_multiple(t1,t2) :
        ''' Return true if exists an integer k such that t1 = nt2 '''
        if t1 == vec_zero : return True
        elif t2 == vec_zero : return False
        else :
            check = 0
            for i in range(len(t1)) :
                if t1[i] == 0 :
                    if t2[i] == 0 : continue
                    else : return False
                elif t2[i] == 0 : return False
                
                if t1[i] % t2[i] != 0 : return False
                if not check : check = (t1[i] // t2[i])
                elif not check == (t1[i] // t2[i]) : return False
            return True
            
    
    def is_included(c1,c2) :
        ''' Check if c1, a constituing sets of a parikh image are is 
        included in c2, an other constituing set '''
        if c1[0] == c2[0] : pass
        else :
            for rep in c2[1] :
                if check_tuple_is_multiple(c1[0],rep) : break
            else :
                return False

        for rep1 in c1[1] :
            for rep2 in c2[1] :
                if check_tuple_is_multiple(rep1,rep2) : break
            else :
                return False
            
        return True
            
    
    def reduce_parikh(P) :
        for couple in P :
            ''' Remove epsilon from repeated part '''
            if vec_zero in couple[1] : couple[1].remove(vec_zero)
            ''' Check if any rep part is a multiple of another rep part '''
            for rep1,rep2 in product(couple[1],couple[1]) :
                if rep1 != rep2 and check_tuple_is_multiple(rep1,rep2) :
                    couple[1].remove(rep1)
        ''' Check if a constituing set of the parikh image is included in 
        an other '''
        dec = 0
        for i in range(len(P)) :
            c1 = P[i-dec]
            for (j,c2) in enumerate(P) :
                if i-dec != j and is_included(c1,c2) :
                    P.remove(c1)
                    dec += 1
                    break;
    
    # Initialisation 
    parikh_im = [[([ (new_tuple_trans(t),set())
                     for t in access_g[s_i][s_o] ] +
                   ([ (vec_zero,set()) ] if s_i == s_o else []) )
                  for s_o in range(len(indexes)) ]
                 for s_i in range(len(indexes)) ]

    # Iteration
    for k in range(len(indexes)) :
        new_parikh = [ [ []
                         for s_o in range(len(indexes)) ]
                       for s_i in range(len(indexes)) ]
    
        for q,l in product(range(len(indexes)),repeat = 2) :
            Lkk_star = (vec_zero,reduce(lambda x,y : x | {y[0]} | y[1],
                                          parikh_im[k][k],set()) )
            Lql = parikh_im[q][l]; Lqk = parikh_im[q][k]; Lkl = parikh_im[k][l]

            def concat_parikh_Lkk(a,b) :
                base = tuple( x + y for x,y in zip(a[0],b[0]) )
                repeat = a[1] | Lkk_star[1] | b[1]
                return (base,repeat)
            
            new_parikh[q][l] = ( Lql +
                                 [ concat_parikh_Lkk(a,b)
                                   for a,b in product(Lqk,Lkl) ] )
            #reduce_parikh(new_parikh[q][l])
            
        parikh_im = new_parikh

    return indexes,trans_list,parikh_im

def parikh_pcal_from_ota(ota,silent = '',inter_pcal = None) :
    states = ota.states; guards = ota.guards; n = len(states); m = len(guards)
    if not inter_pcal : inter_pcal = inter_pcal_from_ota(ota,silent)
    indexes,trans_list,parikh_im = parikh_image(ota,silent)

    e_trans = ota.get_silent_trans(silent)

    def G(i_trans) :
        ''' Given a transition (g,e,g') return G_{e,g} '''
        g,e,g1 = trans_list[i_trans]
        q,lg,hg,op,l = e
        
        return IntervalFromSL(max(ota.guards[lg],ota.guards[g]),
                              ota.guards[hg]) if not op else milib.ZERO

    def G_tuple(t) :
        ''' Given a transition vector return the sum of each of them '''
        return sum( (t[i]*G(i) for i in range(len(trans_list)) if t[i]) ,
                    milib.ZERO)
    
    def is_case_covers(q,mu,l,nu) :
        ''' Is there an repeated component of the parikh image of L(q,l)^nu
        such that a transition with non nul G_e,g appears '''
        iq = indexes.index((q,mu)); il = indexes.index((l,nu))
        P = parikh_im[iq][il]
        for p in P :
            for rep in p[1] :
                res_e = lambda t : not t[1][1][3] and not t[1][2]
                for i,t in filter(res_e,enumerate(trans_list)) :
                    if rep[i] and G(i).vol() > 0 : return (p,rep,i)
        return None

    def compute_alpha_min(p,rep,trans) :
        K0 = G_tuple(p[0])
        Krep = G_tuple(rep)
        assert Krep.vol() > 0
        K = lambda n : K0 + n*Krep

        if K0.vol() == milib.INFINITY : return K0.inf()
        elif Krep.vol() == milib.INFINITY : return (K0 + Krep).inf()
        
        N = int ( (Krep.sup() - K0.vol()) // Krep.vol() ) + 1
        N = max(0,N)
        return K(N).inf()

    def compute_in_case_covers(q,mu,l,nu,alpha) :
        iq = indexes.index((q,mu)); il = indexes.index((l,nu))
        P = parikh_im[iq][il]

        J = ( lambda p,vals :
              ( G_tuple(p[0]) +
                sum( (vals[i]*G_tuple(rep) for i,rep in enumerate(p[1])),
                     milib.ZERO ) ) )
        
        J_res = IntervalFromSL((alpha,True))
        for p in P :
            J_res = J_res | G_tuple(p[0])

        def gen_all_infnorm_M(M,size) :
            return [t1+(M,)*(j+1)+t2 for j in range(size) for i in range(size-j) 
                    for t1 in product(range(M),repeat=i)
                    for t2 in product(range(M),repeat=size-i-j-1) ]
            
        M = 1
        fi = open('log_compute_covers.txt','w')
        while True :
            new_J = J_res
            for p in P :
                for vals in gen_all_infnorm_M(M,len(p[1])) :
                    new_J = J_res | J(p,vals)

            if new_J == J_res : break

            J_res = new_J
            M += 1
            fi.write(str(M))
            fi.flush()

            if M > 10*max(alpha,1) :
                print('Stopped computing in case covers',
                      'reached %i for alpha : %i' % (M,int(alpha)))
                break;
        fi.close()
        return -J_res
                
    def compute_in_case_repeat(q,mu,l,nu) :
        iq = indexes.index((q,mu)); il = indexes.index((l,nu))
        P = parikh_im[iq][il]

        Jegi_list = []
        for p in P :
            J0 = G_tuple(p[0])

            gj_l = []
            for rep in p[1] :
                g_rep = G_tuple(rep)
                assert g_rep.vol() == 0
                g_rep = int(g_rep.inf())
                if g_rep : gj_l.append(g_rep)

            if gj_l :
                nb_gj = len(gj_l)
                lcm_gj = int(reduce(lambda x,y : ruilib.lcm(x,y),gj_l,1))
                prod = [ range(lcm_gj // gj) for gj in gj_l ]
                Jrep = milib.EMPTYSET
                for rj_l in product(*prod) :
                    J_rep = Jrep | (J0 + sum( rj*gj
                                              for rj,gj in zip(rj_l,gj_l)))
                Jegi_list.append(RegUInterval(Irep = -J_rep,period = lcm_gj))
            else :
                Jegi_list.append(-J0)
        return bigor(Jegi_list,RegUInterval())

    parikh_pcal = [ [ [ [] for nu in guards ] for l in states ] for q in states]
    for q,l,nu in product(range(n),range(n),range(m-1)) :
        iq = indexes.index((q,0)); il = indexes.index((l,nu))
        res_e = lambda t : not t[1][1][3] and not t[1][2]
        for i,t in filter(res_e,enumerate(trans_list)) :
            g,e,g1 = t
            assert g1 == 0
            qe,lge,hge,ope,le = e
            for h in inter_pcal[q][qe][g] :
                res = is_case_covers(le,g1,l,nu)
                if res :
                    Jeg = compute_in_case_covers(le,g1,l,nu,
                                                 compute_alpha_min(*res))
                else :
                    Jeg = compute_in_case_repeat(le,g1,l,nu)
                if Jeg : parikh_pcal[q][l][nu].append((h,G(i),Jeg))
    return parikh_pcal

code = '''from Test_sample.{0} import ota
from TILib.milib import Interval, MultiInterval, INFINITY
from TILib.ruilib import RegUInterval
from ATDLib.diagotaumata import DiagOTA

states = ota.states; guards = ota.guards; silent = ota.silent
init_state = ota.init_state

inter_pcal = {1}

parikh_pcal = {2}

trans = {3}

dota = DiagOTA(states,guards,trans,init_state,inter_pcal,parikh_pcal,{4})
'''

def dota_from_ota(ota,retres = False) :
    n = len(ota.states); m = len(ota.guards)
    
    trans = { l : [ [ [None]*n, [None]*n ] for q in ota.states ]
              for l in ota.trans }

    for l in ota.trans :
        for t in ota.trans[l] :
            trans[l][t[4]][int(t[3])][t[0]] = (t[1],t[2])
            
    if retres :
        return DiagOTA(ota.states,ota.guards,trans,ota.init_states,
                       inter_pcal_from_ota(ota),
                       parikh_pcal_from_ota(ota),
                       'diag_' + ota.name)

    def str_MI(I) :
        return 'MultiInterval(' + str(I.sl) + ')' 

    def str_RUI(R) :
        return ','.join('RegUInterval(' +
                str_MI(R.Ibase),
                str_MI(R.Irep),
                str(period),
                str(Irep_dec),
                str(dec),
                "'" + precision + "'")
    
    inter_pcal = inter_pcal_from_ota(ota)
    '''for q,l,nu in product(range(n),range(n),range(m)) :
        res = []
        for g in inter_pcal[q][l][nu] :
            res.append(str_MI(g))
        inter_pcal[q][l][nu] = res'''

    parikh_pcal = parikh_pcal_from_ota(ota)
    '''for q,l,nu in product(range(n),range(n),range(m)) :
        res = []
        for T in parikh_pcal[q][l][nu] :
            res.append((str_MI(T[0]), str_MI(T[1]),
                        str_RUI(T[2]) if T[2].Irep else str_MI(T[2].Ibase)))
        parikh_pcal[q][l][nu] = res'''
        
    wota = open('Test_sample/dota_' + ota.name + '.py','w')                       
    wota.write(code.format(ota.name,
                           inter_pcal,
                           parikh_pcal,
                           trans,
                           repr('diag_' + ota.name)) )
