################################################################################
#           Library for one-clock timed automata simulation                    #
#                                                                              #
#                       JAZIRI Samy - 01/04/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#                                                                              #
################################################################################

from TILib.timlib import *
from TILib.milib import IntervalFromSL, IntervalToSL

from random import randint
from bisect import insort

################################################################################
#### Class OTAutomata
################################################################################

class OTAutomata :
    '''
    '''
    # Class Attribute
    counter = 0
    
    # Constructor
    def __init__(self,states,guards,trans,init_state,name = '',silent = 'e') :
        self.states = states; self.guards = guards; self.trans = trans;
        self.init_state = init_state
        self.silent = 'e'
        if name :
            self.name = name 
        else :
            self.name = 'One-Clock Timed Automata #%i' % OTAutomata.counter
            OTAutomata.counter += 1
        self.access_g = None # Store the access_graph here if already computed
            
    # Methods
    def apply(self,t,tim) :
        ''' Apply the transition t on tim. t = (q,g-,g+,op,q',l)
        with q and q' are the indexes of the states.
        g- and g+ are the indexes of the guards. 
        l is the label.
        op is a boolean : True for Id and False for Reset '''
        g = self.guards
        tim.apply_dec()
        if tim[t[0]] & (t[1],t[2]) :
            if t[3] : tim[t[0]],tim[t[4]] = EMPTYFUTI(g),tim[t[0]]
            else : tim[t[0]],tim[t[4]] = EMPTYFUTI(g),ZEROFUTI(g)
            return True
        else : return False
    
    def action_random(self,l,tim) :
        ''' Simulate the a random transition with label l  
        on the timed interval marking tim '''
        T = self.trans[l][:]
        while T :
            i = randint(0,len(T)-1)
            t = T.pop(i)
            if self.apply(t,tim) : return True
        
        #print('No transition %s enabled' % l)
        return False

    def init_config(self) :
        init_futi = []
        for i in range(0,len(self.states)) :
            init_futi += ( [ZEROFUTI(self.guards)] if i == self.init_state
                           else [EMPTYFUTI(self.guards)] )
        return TIMarking(self.states,self.guards, init_futi)

    def get_silent_trans(self,label = '') :
        if not label : label = self.silent
        return self.trans.get(label,[])
    
    def e_access_graph(self,e_label = '') :

        if not self.access_g or self.access_g[0] != e_label :
            indexes = [ (q,low_g,high_g)
                        for q in range(len(self.states))
                        for low_g in range(len(self.guards)-1)
                        for high_g in range(1,len(self.guards)) ]
            
            e_trans = self.get_silent_trans(e_label)
            
            def si2so(s_i,s_o,t) :
                return (t[0] == s_i[0] and t[4] == s_o[0] and
                        s_i[1] < t[2] and
                        ( ( t[3] and
                            max(t[1],s_i[1]) == s_o[1] and
                            min(t[2],s_i[2]) == s_o[2] ) or
                          ( not t[3] and
                            s_o[1] == 0 and
                            s_o[2] == len(self.guards)-1 ) ) )
        
            self.access_g = (e_label,indexes,
                             [ [ [ (s_i[1],s_i[2],t,s_o[1],s_o[2])
                                   for t in filter(lambda t : si2so(s_i,s_o,t),
                                                   e_trans) ]
                                 for s_o in indexes ]
                               for s_i in indexes ] )        

        return (self.access_g[1][:],
                [ [ self.access_g[2][s_i][s_o][:]
                    for s_o in range(len(self.access_g[1])) ]
                  for s_i in range(len(self.access_g[1])) ] )        


    def __call__(self,l,tim) :
        if type(l) == type('') :
            return self.action_random(l,tim)
        else :
            tim + l
            
 

################################################################################
#### Load OTAutomata from file
################################################################################

code = '''from ATDLib.otaumata import OTAutomata
from TILib.milib import INFINITY
from decimal import Decimal

# States
states = {0} 

# Guards 
guards = {1} + [(INFINITY,False)]


# Transitions 
trans = {2}

# OTAumata
ota = OTAutomata(states,guards,trans,0,'{3}') 
'''

def fload_ota(fota,retres = False) :
    states = []
    init_state = 0
    guards = [(0,False)]
    trans = {}
    for line in fota :
        line.strip()
        
        if line.startswith('#') or (not line) : continue
        
        if not states :
            states = list(map(str.strip,line.split(';')))
            continue

        tstr = list(map(str.strip,line.split(';')))
        t = [None]*5
        t[0] = states.index(tstr[0]); t[4] = states.index(tstr[3])
        t[3] = (not tstr[2] == '0')

        t[1],t[2] = (IntervalToSL(tstr[1]) if tstr[1]
                     else ( (0,False),(milib.INFINITY,False) ) )
        for i in [1,2] :
            if not (t[i] in guards) and (not t[i] == (milib.INFINITY,False)) :
                insort(guards,t[i])

        trans[tstr[4]] = trans.get(tstr[4],[]) + [t]

    for label in trans :
        new_label = []
        for t in trans[label] :
            for i in [1,2] :
                if t[i] == (milib.INFINITY,False) : t[i] = len(guards)
                else : t[i] = guards.index(t[i])
            new_label.append(tuple(t))
        trans[label] = new_label

    name = fota.name.split('.')[-2].split('/')[-1]
        
    wota = open('Files/'+name + '.py','w')
    wota.write(code.format(states,guards,trans,name))
    wota.close()

    if retres :
        guards += [(milib.INFINITY,False)]
        return OTAutomata(states,guards,trans,0,name)
    
    return
