################################################################################
#                   Library for multi-interval calculus                        #
#                                                                              #
#                       JAZIRI Samy - 30/03/2018                               #
#                   Mail : samy.jaziri@ens-cachan.fr                           #
#                                                                              #
#                                                                              #
# Notes :                                                                      #
#  - We represent a semi-line by a pairs (v,b) with v a float and b a boolean. #
#    v is the value of the lower bound and b is True iff the semi-line is open.#
#    (3,True) stands then for ]3,inf[, and (12,False) for [12,inf[             #
#  - An interval is represented by two semi-line l and r, and is obtain by     #
#    intersecting the first one by the complement of the other.                #
#    Hence ]3,12] is represented by ]3,inf[ \cap \overline{]12,inf[}           #
#    i.e. by (3,True) and (12,true)                                            #
#  - A finite union of interval (multi-interval) is then a list of intervals   #
#    Multi-intervals subsumes semi-lines and intervals, so we create a class   #
#    only to represent multi-intervals.                                        #
#                                                                              #
################################################################################

from bisect import bisect, bisect_left
from functools import reduce
from decimal import Decimal

INFINITY = Decimal('inf')

################################################################################
#### Class MultiInterval
################################################################################

class MultiInterval :
    ''' 
    Finite union of intervals
    Mutable sequence
    
    Note : 
       We implement it by a list of an even number of semiline : 
       even position are left bounds and odd position right bounds. 
       The sequence is strictly increasing in such way that our finite union of
       interval U always verify :
       
       - Two intervals in U are not overlapping, and not adjacent. 
         ([2,3[ and [3,4[ are not overlapping but adjacent intervals)  
       - An interval in U is not empty.
       - All intervals in U are ordered from the smallest to the biggest values

    Instantiate the emptyset by default.

    Args :
       sl : a sorted list of semi-lines, of even length

    Attributes :
       sl : a sorted list of semi-line, of even length.

    '''
    
    # Constructor
    def __init__(self, l = [], precision = '.2') :
        ''' Instanciate a multi-interval from the list l of semi-lines.
        By default the emptyset. '''
        self.sl = l
        self.pf = '%' + precision + 'f'
        self.Ibase = self
        self.Irep = None
        self.Irep_dec = self.inf()-1
        self.period = 1
        self.dec = 0

    # Representation
    def __str__(self) :
        ''' Return a sting equivalent to the code generating the object '''
        if not self : return 'EMPTYSET'

        def reprinterval(I) :
            ''' Take two semi-lines a and b, and return a string to represent
            the interval they define '''
            l = [ ']' if x[1] else '[' for x in I.sl ]
            return (l[0] + self.pf % I.sl[0][0] + ',' +
                    self.pf % I.sl[1][0] + l[1] )
        
        return ' U '.join(map(reprinterval,self))

    def __repr__(self) :
        ''' Return a string with the representation of self '''
        return 'MultiInterval('+str(self.sl).replace('inf','INFINITY')+')'
    
    # Methods
    def vol(self) :
        ''' Return the volume of self. 
        I.e. the sum of all length of the intervals in self. '''
        if self.sl[0][0] == INFINITY or self.sl[-1][0] == INFINITY :
            return INFINITY
        return sum(b[0]-a[0] for a,b in zip(self.sl[::2],self.sl[1::2]))

    def sup(self) :
        ''' Return the upper bound of the multi-interval '''
        if not self : return -INFINITY
        return self.sl[-1][0]

    def sup_is_max(self) :
        ''' Return a boolean saying if the sup of the multi interval is a max :
        True if its a max, False other wise '''
        if not self : return False
        return self.sl[-1][1]
    
    def inf(self) :
        ''' Return the lower bound of the multi-interval '''
        if not self : return INFINITY
        return self.sl[0][0]

    def inf_is_min(self) : 
        ''' Return a boolean saying if the inf of the multi interval is a min :
        True if its a min, False other wise '''
        if not self : return False
        return self.sl[0][1]

    def apply_dec(self,padding = 0) : self = self + padding
    
    # Operator Overloading
    #### Comparison
    def __eq__(self,other) :
        ''' self == other '''
        return self.sl == other.sl

    def __ne__(self,other) :
        ''' self != other '''
        return self.sl != other.sl        
    #### Sequence Operators
    def __len__(self) :
        ''' The number of intervals in self '''
        return len(self.sl) // 2

    def __getitem__(self,i) :
        ''' Return the i^th interval in the multi-interval, 
        in the form (l,r) with l and r two semi-lines.  '''
        if isinstance(i,slice) :
            start = i.start*2 if i.start else None
            stop = i.stop*2 if i.start else None
            return MultiInterval(self.sl[start:stop])
        return MultiInterval([self.sl[i*2],self.sl[i*2+1]])

    def __setitem__(self,i,I) :
        ''' Set the i^th interval in the multi-interval to the interval I, 
        in the form (l,r) with l and r the two semi-lines of I.  '''
        self.sl[i*2] = I.sl[0]; self.sl[i*2+1] = I.sl[1]
    
    #### Set Operators
    def __and__(self,other) :
        ''' Return the intersection of self and other.
        Complexity : O( |other|*log(|self|) + (|other|-1)*|self| )
        If other is an interval the complexity is hence : O( log(|self|) ) '''
        if (not self) | (not other) : return EMPTYSET
    
        def interinterval(I) :
            ''' Take two semi-lines a and b, and return the intersection 
            of self with the interval they define. 
            Complexity : O( log(|self|) ) '''
            l_pos = bisect(self.sl,I.sl[0])
            r_pos = bisect_left(self.sl,I.sl[1],l_pos)
            return MultiInterval( ([I.sl[0]] if l_pos % 2 else [] ) +
                                  self.sl[l_pos:r_pos] +
                                  ([I.sl[1]] if r_pos % 2 else [] ) )

        return bigor(map(interinterval,other))

    def __or__(self,other) :
        ''' Return the union of self and other.
        Complexity : O( |self|+|other| )'''
        if not other : return self
        if not self : return other

        if not isinstance(other,MultiInterval) :
            return other | self
        
        pos_s = pos_o = 0
        res_sl = []
        while True :
            if (pos_s % 2) :
                if self.sl[pos_s] <= other.sl[pos_o] :
                    res_sl.append(self.sl[pos_s])
                    pos_s += 1
                else :
                    pos_o += 1
                    if self.sl[pos_s] <= other.sl[pos_o] : pos_s += 1
                    else : pos_o += 1
            elif (pos_o % 2) :
                if other.sl[pos_o] <= self.sl[pos_s] :
                    res_sl.append(other.sl[pos_o])
                    pos_o += 1
                else :
                    pos_s += 1
                    if other.sl[pos_o] <= self.sl[pos_s] : pos_o += 1
                    else : pos_s += 1
            else :
                if self.sl[pos_s] <= other.sl[pos_o] :
                    res_sl.append(self.sl[pos_s])
                    pos_s += 1
                else :
                    res_sl.append(other.sl[pos_o])
                    pos_o += 1

            if pos_s == len(self.sl) : res_sl += other.sl[pos_o:]; break
            if pos_o == len(other.sl) : res_sl += self.sl[pos_s:]; break

        return MultiInterval(res_sl)

    def __invert__(self) :
        ''' Return the complement of self.
        Complexity : O(1) '''
        if not self : return Interval(']-inf,inf[')
        res_sl = ([] if self.sl[0] == (-INFINITY,True)
                  else [(-INFINITY,True),self.sl[0]] )
        res_sl += self.sl[1:-1]
        res_sl += ([] if self.sl[-1] == (INFINITY,False)
                  else [self.sl[-1],(INFINITY,False)] )
        return MultiInterval(res_sl)
                 

    #### Arithmetic Operators
    def __neg__(self) :
        ''' Return -self 
        Complexity : O( |self| ) '''
        return MultiInterval([ (-a,not b) for a,b in reversed(self.sl) ])
    
    def __add__(self,other) :
        ''' Return the addition of self and other.
        Complexity : O( |other|*|self| + (|other|-1)*|self| )
        If other is an interval the complexity is hence : O( |self| ) '''

        def add_pair(a,b) :
            ''' Given two semi-line, make the addition.
            i.e. the addition of their value and the logical and of booleans '''
            return (a[0]+b[0],a[1] and b[1])

        if (type(other) == type(0.0) or
            type(other) == type(0) or
            type(other) == type(Decimal(0)) ) :
            # Complexity : O(|self|)
            return MultiInterval( [add_pair(a,(other,True)) for a in self.sl] )

        
        if (not self) | (not other) : return EMPTYSET
        
        def addinterval(I) : 
            ''' Take two semi-lines a and b, and return the additions
            of self with the interval they define. 
            Complexity : O( |self| ) '''
            res_sl = [ add_pair(self.sl[0],I.sl[0]) ]
            for x,y in ( (add_pair(a,I.sl[1]), add_pair(b,I.sl[0]))
                        for a,b in zip(self.sl[1:-1:2],self.sl[2::2]) ) :
                res_sl += [x,y] if x < y else []
            res_sl += [ add_pair(self.sl[-1],I.sl[1]) ]
            return MultiInterval( res_sl )

        return bigor(list(map(addinterval,other)))

    def __sub__(self,other) :
        ''' Return the subtraction of self and other.
        Complexity : O( |other|*|self| + (|other|-1)*|self| )
        If other is an interval the complexity is hence : O( |self| ) '''
        def sub_pair(a,b) :
            ''' Given two semi-line, make the addition.
            i.e. the addition of their value and the logical and of booleans '''
            return (a[0]-b[0],a[1] and not b[1])

        if (type(other) == type(0.0) or
            type(other) == type(0) or
            type(other) == type(Decimal(0)) ) :
            # Complexity : O(|self|)
            return MultiInterval( [sub_pair(a,(other,False)) for a in self.sl] )

        if (not self) | (not other) : return EMPTYSET

        
        def subinterval(I) : 
            ''' Take two semi-lines a and b, and return the subtraction
            of self with the interval they define. 
            Complexity : O( |self| ) '''
            res_sl = [ sub_pair(self.sl[0],I.sl[1]) ]
            for x,y in ( (sub_pair(a,I.sl[0]), sub_pair(b,I.sl[1]))
                        for a,b in zip(self.sl[1:-1:2],self.sl[2::2]) ) :
                res_sl += [x,y] if x < y else []
            res_sl += [ sub_pair(self.sl[-1],I.sl[0]) ]
            return MultiInterval( res_sl )

        return bigor(list(map(subinterval,other)))


    def __rmul__(self,n) :
        ''' Return the multiplication of self by a float n.
        Complexity : O( |self| )'''
        assert n >= 0
        return MultiInterval([(n*x,b) for (x,b) in self.sl])

################################################################################
#### MultiInterval Functions
################################################################################

def IntervalFromSL(a = (-INFINITY,True) , b = (INFINITY,False)) :
    ''' Return the multi-interval with only one interval defined by 
    the two semi-lines a and b '''
    if a >= b : return EMPTYSET
    return MultiInterval([a,b])

def IntervalToSL(s) :
    I = Interval(s)
    return (s.sl[0],s.sl[1])

def Interval(s) :
    ''' Return the interval defined by the string s '''
    s1,s2 = s.split(',')
    l = (Decimal(s1[1:]),s[0] == ']')
    r = (Decimal(s2[:-1]),s[-1] == ']')
    assert l < r
    return MultiInterval([l,r])

def IntervalToSL(s) :
    ''' Return the interval defined by the string s in the form of two
    semi-lines l and r'''
    s1,s2 = s.split(',')
    l = (Decimal(s1[1:]),s[0] == ']')
    r = (Decimal(s2[:-1]),s[-1] == ']')
    assert l < r
    return (l,r)

EMPTYSET = MultiInterval()
ZERO = Interval('[0,0]')
NREALS = Interval('[-inf,0]')
REALS = Interval('[-inf,inf]')

def bigor(lmi,init = EMPTYSET) :
    ''' Take a list of objects and return the or of all its 
    elements. '''
    return reduce((lambda x,y : x | y),lmi,init)

################################################################################
#### Test Functions
################################################################################

if __name__ == '__main__' :

    from random import randint, uniform, sample
    from sys import stdout
    
    #Membership definition
    def bel(x,U) :
        ''' Return True if x belongs (set membership) to the multi-interval U. 
        This is done by testing U \cap [x,x] == EMPTYSET '''
        return bool(U & Interval('[{0},{0}]'.format(x)))
    
    #Membership test function
    def test_membership(sample) :
        print( 'Membership correction.\n' + 
               'Correction of all tests are based on membership correction' )

        progress = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            for I1,I2 in zip(U,MultiInterval(U.sl[1:-1])) :
                halfI1 = (I1.sl[1][0]+I1.sl[0][0]) / 2
                halfI2 = (I2.sl[1][0]+I2.sl[0][0]) / 2
                if not bel( halfI1, U) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error IN on MultiInterval {0} ' +
                            '\twith element {1}').format(U,halfI1) )
                    break;
                elif bel( halfI2, U) : 
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error NOTIN on MultiInterval {0} ' +
                            '\twith element {1}').format(U,halfI2) )
                    break
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                continue
            break
        else :
            print('] %i tests succeeded' % NUMBER)
            print('O == Membership correction test succeeded\n')

    #Intersection test function
    def test_intersection(sample) :
        print( 'Intersection tests.\n' +
               'Intersection by intervals (50%) and multi interval (50%)' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            V = sample[randint(0,SIZE-1)]
            
            (a,b,c,d) = ((randint(0,SIZE*1000) / 100),randint(0,1),
                         (randint(0,SIZE*1000) / 100),randint(0,1) )
            (a,b),(c,d) = min((a,b),(c,d)),max((a,b),(c,d))
            if (a,b) == (c,d) : c = c+0.5
            
            I = IntervalFromSL((a,b),(c,d))
            W1 = U & I
            W2 = U & V

            count1 = 0
            for j in range(0,VERIFNUMBER) :
                x = uniform(0,SIZE*10)
                y = uniform(0,SIZE*10)
                count1 += 1 if bel(x,W1) else 0
                if (bel(x,U) and bel(x,I)) != bel(x,W1) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error InI on MultiIntervals {0} and {1} ' +
                            '\twith element {2}').format(U,I,x) )
                    break;
                if (bel(y,U) and bel(y,V)) != bel(y,W2) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error InM on MultiIntervals {0} and {1} ' +
                            '\twith element {2}').format(U,V,y) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % (NUMBER*2) )
            print('O == Intersection correction test succeeded',end='')
            print(' Average x in U \cap V is %i/%i \n' %
                  (count // NUMBER, VERIFNUMBER) )

    #Union test function
    def test_union(sample) :
        print( 'Union tests.\n' +
               'Union between multi interval' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            V = sample[randint(0,SIZE-1)]            
            W = U | V

            count1 = 0
            for j in range(0,VERIFNUMBER) :
                x = uniform(0,SIZE*10)
                count1 += 1 if bel(x,W) else 0
                if (bel(x,U) or bel(x,V)) != bel(x,W) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error UnM on MultiIntervals {0} and {1} ' +
                            '\twith element {2}').format(U,V,x) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % (NUMBER*2))
            print('O == Union correction test succeeded',end='')
            print(' Average x in U \cup V is %i/%i \n' %
                  (count // NUMBER, VERIFNUMBER) )

    #Complement test function
    def test_complementation(sample) :
        print( 'Complementation tests.\n' +
               'Complementation of a multi interval' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            W = ~U

            count1 = 0
            for j in range(0,VERIFNUMBER) :
                x = uniform(0,SIZE*10)
                count1 += 1 if bel(x,W) else 0
                if (bel(x,U)) == bel(x,W) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error CoM on MultiIntervals {0} ' +
                            '\twith element {1}').format(U,x) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % NUMBER)
            print('O == Complementation correction test succeeded',end='')
            print(' Average x in ~U is %i/%i \n' % (count // NUMBER, VERIFNUMBER) )

    #Addition test function
    def test_addition(sample) :
        print( 'Addition tests.\n' +
               'Addition by numbers (50%), intervals (25%) ' +
               'and multi interval (25%)' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            if i % 4 == 1 :
                V = sample[randint(0,SIZE-1)]
                W = U + V
                if V :
                    I = V[randint(0,len(V)-1)]
                    d = (I.sl[0][0] + I.sl[1][0]) / 2
                else :
                    if W :
                        print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error AddIES ' +
                                'on MultiIntervals \n {1} \n and {2}' +
                            '').format(U,W) )
                        break
                    else :
                        continue
                        
            elif i % 4 == 3 :
                (a,b,c,d) = ((randint(0,SIZE*1000) / 100),randint(0,1),
                             (randint(0,SIZE*1000) / 100),randint(0,1) )
                (a,b),(c,d) = min((a,b),(c,d)),max((a,b),(c,d))
                if (a,b) == (c,d) : c = c+0.5
                I = IntervalFromSL((a,b),(c,d))
                W = U + I
                d = (I.sl[0][0] + I.sl[1][0]) / 2
            else :
                d = randint(0,SIZE*1000) / 100
                W = U + d

            count1 = 0
            for j in range(0,VERIFNUMBER*2) :
                x = randint(0,SIZE*1000) / 100
                count1 += 1 if bel(x,W) else 0
                if not bel(x+d,W) and bel(x,U) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error AddI{0}-{4} ' +
                            'on MultiIntervals \n {1} \n and {5}' +
                            '\twith element {2} and padding {3}' +
                            '').format(i%4,U,x,d,bel(x,W),W) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % NUMBER)
            print('O == Addition correction test succeeded',end='')
            print(' Average x in U + V is %i/%i '
                  % (count // (2*NUMBER), VERIFNUMBER) )

    #Subtraction test function
    def test_subtraction(sample) :
        print( 'Subtraction tests.\n' +
               'Subtraction by multiinterval (50%), intervals (25%) ' +
               'and numbers (25%)' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            if (i % 4) % 2 == 0 :
                V = sample[randint(0,SIZE-1)]
                W = U - V
                if V :
                    I = V[randint(0,len(V)-1)]
                    d = (I.sl[0][0] + I.sl[1][0]) / 2
                else :
                    if W :
                        print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error AddIES ' +
                                'on MultiIntervals \n {1} \n and {2}' +
                            '').format(U,W) )
                        break
                    else :
                        continue
                
            elif i % 4 == 3 :
                (a,b,c,d) = ((randint(0,SIZE*1000) / 100),randint(0,1),
                             (randint(0,SIZE*1000) / 100),randint(0,1) )
                (a,b),(c,d) = min((a,b),(c,d)),max((a,b),(c,d))
                if (a,b) == (c,d) : c = c+0.5
                I = IntervalFromSL((a,b),(c,d))
                W = U - I
                d = (I.sl[0][0] + I.sl[1][0]) / 2
            else :
                d = randint(0,SIZE*1000) / 100
                W = U - d

            count1 = 0
            for j in range(0,VERIFNUMBER*2) :
                x = uniform(0,SIZE*10)
                count1 += 1 if bel(x,W) else 0
                if bel(x,U) and not bel(x-d,W) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error SubI{0}-{4} ' +
                            'on MultiIntervals \n {1} \n and {5}' +
                            '\twith element {2} and padding {3}' +
                            '').format(i%4,U,x,d,bel(x,W),W) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % NUMBER)
            print('O == Subtraction correction test succeeded',end='')
            print(' Average x in U-V is %i/%i \n'
                  % (count // 2*NUMBER, VERIFNUMBER) )

    #Multiplication test function
    def test_multiplication(sample) :
        print( 'Multiplication tests.\n' +
               'Multiplication by an integer' )
        
        progress = 0
        count = 0
        print('[',end='')
        for (i,U) in enumerate(sample) :
            n = randint(0,SIZE)
            W = n*U 

            count1 = 0
            for j in range(0,VERIFNUMBER) :
                x = uniform(0,SIZE*10)
                count1 += 1 if bel(x,W) else 0
                if bel(x,U) and not bel(n*x,W) :
                    print( ('>>>'*(10-progress) + ']\n' +
                            'X == Error MulI{0} ' +
                            'on MultiIntervals \n {1} \n and {4}' +
                            '\twith element {2} and integer {3}' +
                            '').format(bel(x,W),U,x,n,W) )
                    break;
            else : 
                if (i*10 // NUMBER) >= progress :
                    print('>>>',end='')
                    stdout.flush()
                    progress += 1
                count = count + count1
                continue
            break
        else :
            print('] %i tests succeeded' % NUMBER)
            print('O == Multiplication correction test succeeded',end='')
            print(' Average x in n*U is %i/%i \n' % (count // NUMBER, VERIFNUMBER) )

    #Running tests
    def run(l) :
        ''' Take a list of keys and make the corresponding tests '''
        tests = {
            #Membership correction
            0 : test_membership,
            #Intersection correction
            'i' : test_intersection,
            #Union correction
            'u' : test_union,
            #Complementation correction
            'c' : test_complementation,
            #Addition correction
            'a' : test_addition,
            #Subtraction correction
            's' : test_subtraction,
            #Multiplication correction
            'm' : test_multiplication,
        }

        for k in l :
            tests[k](sample)


################################################################################
#### Running Tests
################################################################################

    print( 'Tests :\n' + '-'*30 )

    NUMBER = 1000
    SIZE = 100
    VERIFNUMBER = 1000
    # Sample Generation
    print((('Sample generation. %i samples of size <= %i.\n' % (NUMBER,SIZE)) +
           'Tests verified with %i values.' % VERIFNUMBER))

    population = ([(x / 100,True) for x in range(SIZE*1000)] +
                  [(x / 100,False) for x in range(SIZE*1000)] ) 
    sample = [ MultiInterval(sorted(sample(population,randint(0,SIZE)*2)))
               for x in range(NUMBER) ]

    open('sample.txt','w').write(reduce(lambda x,y : x + str(y) + '\n',
                                        sample,''))
    print('-> Done. The sample is stored in sample.txt\n')

    # 0 : Membership correction
    # 'i' : Intersection correction
    # 'u' : Union correction 
    # 'c' : Complementation correction
    # 'a' : Addition correction
    # 's' : Subtraction correction
    # 'm' : Multiplication correction

    #Test list todo
    test_list = [0,'i','a','u']
    #run(test_list)
    print(sample[1])
    print('Tada : ' + str(sample[1][2:9:3]))
